@@ -35,6 +35,17 @@ urlpatterns = patterns(
     url(r'^reports/(?P<name>[a-zA-Z0-9-_]+)/$', 'report_detail'),
     url(r'^tests/$', 'test_list'),
     url(r'^tests/(?P<test_id>[^/]+)/$', 'test_detail'),
+    url(r'^filters/$', 'filters_list'),
+    url(r'^filters/\+add$', 'filter_add'),
+    url(r'^filters/\+add-preview-json$', 'filter_preview_json'),
+    url(r'^filters/\+add-cases-for-test-json$', 'filter_add_cases_for_test_json'),
+    url(r'^filters/\+attribute-name-completion-json$', 'filter_attr_name_completion_json'),
+    url(r'^filters/\+attribute-value-completion-json$', 'filter_attr_value_completion_json'),
+    url(r'^filters/~(?P<username>[a-zA-Z0-9-_]+)/(?P<name>[a-zA-Z0-9-_]+)$', 'filter_detail'),
+    url(r'^filters/~(?P<username>[a-zA-Z0-9-_]+)/(?P<name>[a-zA-Z0-9-_]+)/json$', 'filter_json'),
+    url(r'^filters/~(?P<username>[a-zA-Z0-9-_]+)/(?P<name>[a-zA-Z0-9-_]+)/\+edit$', 'filter_edit'),
+    url(r'^filters/~(?P<username>[a-zA-Z0-9-_]+)/(?P<name>[a-zA-Z0-9-_]+)/\+subscribe$', 'filter_subscribe'),
+    url(r'^filters/~(?P<username>[a-zA-Z0-9-_]+)/(?P<name>[a-zA-Z0-9-_]+)/\+delete$', 'filter_delete'),
     url(r'^xml-rpc/$', linaro_django_xmlrpc.views.handler, 
         name='dashboard_app.views.dashboard_xml_rpc_handler',
         kwargs={
